import unittest

import numpy as np
from numpy.testing import assert_almost_equal

from pyscf import gto, scf
from pyscf.qsdopt.qsd_optimizer import QSD


class QSDOptTests(unittest.TestCase):
    def test_H2O_TS(self):
        mol = gto.M(atom="O 0 0 0; H 0 0 1.2; H 0, 0.1, -1.2", basis="minao", verbose=0)
        mf = scf.RHF(mol)
        optimizer = QSD(mf, stationary_point="TS")
        optimizer.kernel()
        coords = mol.atom_coords()
        v1 = coords[0, :] - coords[1, :]
        v2 = coords[0, :] - coords[2, :]
        d1 = np.linalg.norm(v1)
        d2 = np.linalg.norm(v2)
        phi = np.arccos(v1.dot(v2) / (d1 * d2))
        assert_almost_equal(d1, d2, 4)
        assert_almost_equal(phi, np.pi, 4)
        self.assertTrue(optimizer.converged)

    def test_H2O_min(self):
        mol = gto.M(atom="O 0 0 0; H 0 0 1.2; H 0, 0.5, -1.2", basis="minao", verbose=0)
        mf = scf.RHF(mol)
        optimizer = QSD(mf, stationary_point="min")
        optimizer.kernel()
        coords = mol.atom_coords()
        v1 = coords[0, :] - coords[1, :]
        v2 = coords[0, :] - coords[2, :]
        d1 = np.linalg.norm(v1)
        d2 = np.linalg.norm(v2)
        phi = np.arccos(v1.dot(v2) / (d1 * d2))
        assert_almost_equal(d1, d2, 4)
        assert_almost_equal(phi, np.radians(104.3), 4)
        self.assertTrue(optimizer.converged)



# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from pyscf import lib
from pyscf.grad.rhf import GradientsMixin
from pyscf.qsdopt.hesstools import (
    filter_hessian,
    hess_BFGS_update,
    hess_powell_update,
    numhess,
)


def kernel(
    g_scanner,
    stationary_point,
    hess_update_rule,
    hess_update_freq=0,
    numhess_method="forward",
    max_iter=100,
    step=0.1,
    hmin=1e-6,
    gthres=1e-5,
):
    """
    Quadratic Steepest Descent method for transition state and minima optimization.
    Implemeted following: J. Chem. Phys., 101, 2157, 1994; 10.1063/1.467721
    Arguments:
        hess_update_freq: Iterations before next numerical hessian calculation. If 0 the
            hessian is not updated in the optimization process unless problematic region is reached.
        numhess:method: Method for numerical hessian calculation. "forward" for forward differences
            calculation, "central" for central differences calculation.
        max_iter: Maximum number of iterations in the optimization process.
        step: Maximum step at each optimization iteration.
        hmin: Minimum distance between next point (x1) and quadratic form minima (m) to make x1 = m.
        gthres: Gradient threshold to consider stationary point reached.

    """
    converged = False
    ITA = 0
    ITAM = 10
    m = g_scanner.mol.atom_mass_list()
    sm = np.sqrt(m)
    sm3 = np.repeat(sm, 3)
    nat = len(m)
    x0 = g_scanner.mol.atom_coords().flatten()
    energy, g0 = g_scanner(g_scanner.mol)
    g0 = g0.flatten()
    H = numhess(g_scanner.mol, g_scanner, g0, numhess_method)
    H = filter_hessian(g_scanner.mol, H)
    H = np.einsum("ij, i, j -> ij", H, 1 / sm3, 1 / sm3)
    inc = _qsd_step(x0 * sm3, g0 / sm3, H, sm3, stationary_point, step=step)
    x1 = x0 + inc

    print("Iteration   Energy       Gradient Norm   Increment norm")
    for it in range(1, max_iter):
        x_1 = x0.copy()
        g_1 = g0.copy()
        x0 = x1.copy()
        g_scanner.mol.set_geom_(x0.reshape(nat, 3), unit="Bohr")
        energy, g0 = g_scanner(g_scanner.mol)
        g0 = g0.flatten()
        gnorm = np.linalg.norm(g0)
        incnorm = np.linalg.norm(inc)
        print(f"{it:3d}         {energy:10.7f}  {gnorm:4.2e}        {incnorm:4.2e}")
        if gnorm < gthres or incnorm < hmin or ITA > ITAM:
            converged = True
            break
        if hess_update_freq > 0 and it % hess_update_freq == 0 or ITA >= 4:
            print("Recalculating hessian")
            H = numhess(g_scanner.mol, g_scanner, g0, numhess_method)
        else:
            dH = hess_update_rule(H, x0 - x_1, g0 - g_1)
            H += dH
        Hf = filter_hessian(g_scanner.mol, H)
        Hf = np.einsum("ij, i, j -> ij", Hf, 1 / sm3, 1 / sm3)
        inc = _qsd_step(x0 * sm3, g0 / sm3, Hf, sm3, stationary_point, step=step)
        x1 = x0 + inc

        val = (x_1 - x0) @ (x0 - x1)
        if val > 0.0:
            ITA = 0
        else:
            ITA += 1

    g_scanner.mol.set_geom_(x1.reshape(nat, 3), unit="Bohr")
    return converged, g_scanner.mol


def _qsd_step(x0, g, H, sm3, stationary_point, step=1e-1):
    zero_thres = 1e-10
    step_thres = 1e-1 * step
    eigval, eigvec = np.linalg.eigh(H)
    NMeigvec = eigvec.T[np.abs(eigval) > zero_thres]
    hNM = eigval[np.abs(eigval) > zero_thres]
    gNM = NMeigvec @ g
    delta = gNM / hNM
    if stationary_point == "TS":
        hNM[0] = -hNM[0]

    inc = NMeigvec.T @ delta / sm3
    if np.linalg.norm(inc) < step:
        return -inc

    if all(hNM > 0):
        umin = 0e0
    else:
        umin = 1e-25

    incNM = (umin ** hNM - 1e0) * delta
    inc = NMeigvec.T @ incNM / sm3
    inc_norm = np.linalg.norm(inc)
    inc_max = inc_norm

    if inc_norm > step:
        # Search u value.
        umax = 1e0
        inc_min = 0e0
        while (
            np.abs(inc_norm - step) > step_thres and np.abs(inc_max - inc_min) > 1e-20
        ):
            u = (umax + umin) / 2.0
            incNM = (u ** hNM - 1e0) * delta
            inc = NMeigvec.T @ incNM / sm3
            inc_norm = np.linalg.norm(inc)
            if inc_norm < step:
                umax = u
                inc_min = inc_norm
            else:
                umin = u
                inc_max = inc_norm
    return inc


class QSD(lib.StreamObject):
    """QSD optimizer."""

    def __init__(self, method, stationary_point="min"):
        assert stationary_point in ["min", "TS"]
        self.method = method
        self.stationary_point = stationary_point
        self.converged = False
        if self.stationary_point == "TS":
            self.hess_update = hess_powell_update
        elif self.stationary_point == "min":
            self.hess_update = hess_BFGS_update

    def kernel(self, **kwargs):
        if isinstance(self.method, lib.GradScanner):
            g_scanner = self.method
        elif isinstance(self.method, GradientsMixin):
            g_scanner = self.method.as_scanner()
        elif getattr(self.method, "nuc_grad_method", None):
            g_scanner = self.method.nuc_grad_method().as_scanner()
        else:
            raise NotImplementedError(
                "Nuclear gradients of %s not available" % self.method
            )
        self.converged, self.mol = kernel(
            g_scanner, self.stationary_point, self.hess_update, **kwargs
        )


if __name__ == "__main__":
    from pyscf import gto, scf

    mol = gto.M(atom="O 0 0 0; H 0 0 1.2; H 0, 0.1, -1.2", basis="minao", verbose=0)
    mf = scf.RHF(mol)
    optimizer = QSD(mf, stationary_point="TS")
    optimizer.kernel()
    print(f"Converged: {optimizer.converged}")
    print(mol.atom_coords())



import unittest

import numpy as np
from numpy.testing import assert_almost_equal

from pyscf import gto, scf
from pyscf.qsdopt.hesstools import (
    central_differences_hess,
    filter_hessian,
    forward_differences_hess,
    hess_BFGS_update,
    hess_powell_update,
)

zero_thres = 1e-15


class MyScanner:
    def __call__(self, mol):
        x = mol.atom_coords()
        return x[0, 0] ** 2 * x[0, 1] + x[0, 2] ** 2, np.array(
            [2 * x[0, 0] * x[0, 1], x[0, 0] ** 2, 2 * x[0, 2]]
        )


class HessianTests(unittest.TestCase):
    def test_centralhess(self):
        scanner = MyScanner()
        mol = gto.M(atom="O 1 2 3", basis="minao", unit="Bohr")
        H = central_differences_hess(mol, scanner)
        assert_almost_equal(
            H, np.array([[4.0, 2.0, 0.0], [2.0, 0.0, 0.0], [0.0, 0.0, 2.0]])
        )

    def test_forwardhess(self):
        scanner = MyScanner()
        mol = gto.M(atom="O 1 2 3", basis="minao", unit="Bohr")
        f, g0 = scanner(mol)
        H = forward_differences_hess(mol, scanner, g0)
        assert_almost_equal(
            H, np.array([[4.0, 2.0, 0.0], [2.0, 0.0, 0.0], [0.0, 0.0, 2.0]]), decimal=4
        )

    def test_filterhess1(self):
        mol = gto.M(
            atom="""O 0. 0.40846003 0.60306451
                    H 0. -0.770609 2.20504746
                    H 0. 2.2518751 1.3492855""",
            basis="minao",
            verbose=0,
            unit="Bohr",
        )
        mf = scf.RHF(mol)
        g_scanner = mf.nuc_grad_method().as_scanner()
        H = central_differences_hess(mol, g_scanner)

        H = filter_hessian(mol, H)
        feigval = np.linalg.eigvalsh(H)
        self.assertEqual(len(feigval[feigval < zero_thres]), 6)

    def test_filterhess2(self):
        mol = gto.M(
            atom="H 0 0 0; H 0 0 1.603",
            basis="minao",
            verbose=0,
            unit="Bohr",
        )
        mf = scf.RHF(mol)
        g_scanner = mf.nuc_grad_method().as_scanner()
        H = central_differences_hess(mol, g_scanner)

        H = filter_hessian(mol, H)
        feigval = np.linalg.eigvalsh(H)
        self.assertEqual(len(feigval[feigval < zero_thres]), 5)

    def test_hess_powell_update(self):
        def func(x, y):
            return (
                x ** 3 + x * y ** 2,
                np.array([3 * x ** 2 + y ** 2, 2 * x * y]),
                np.array([[6 * x, 2 * y], [2 * y, 2 * x]]),
            )

        x1 = np.array([0.1, 0.2])
        x2 = np.array([0.1, 0.3])

        f1, g1, H1 = func(x1[0], x1[1])
        f2, g2, H2 = func(x2[0], x2[1])

        dH = hess_powell_update(H1, (x2 - x1), (g2 - g1))
        assert_almost_equal(dH, np.array([[0.0, 1e-1], [1e-1, 0.0]]))

    def test_hess_BFGS_update(self):
        def func(x, y):
            return (
                x ** 3 + x * y ** 2,
                np.array([3 * x ** 2 + y ** 2, 2 * x * y]),
                np.array([[6 * x, 2 * y], [2 * y, 2 * x]]),
            )

        x1 = np.array([0.1, 0.2])
        x2 = np.array([0.1, 0.3])

        f1, g1, H1 = func(x1[0], x1[1])
        f2, g2, H2 = func(x2[0], x2[1])

        dH = hess_BFGS_update(H1, (x2 - x1), (g2 - g1))
        assert_almost_equal(dH, np.array([[4.5e-1, 1e-1], [1e-1, 0.0]]))



# Copyright 2014-2018 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np

from pyscf.hessian.thermo import _get_TR


def numhess(mol, g_scanner, g0, method):
    if method == "forward":
        H = forward_differences_hess(mol, g_scanner, g0)
    elif method == "central":
        H = central_differences_hess(mol, g_scanner)
    return H


def forward_differences_hess(mol, g_scanner, g0):
    """Evaluate numerical hessian of the energy using forward differences."""
    delta = 1e-4
    twodelta = 2 * delta
    geom = mol.atom_coords()
    nat = geom.shape[0]
    ndim = 3 * nat
    H = np.zeros([ndim, ndim])

    for iat in range(nat):
        for icoor in range(3):
            i = 3 * iat + icoor
            _geom = geom.copy()
            _geom[iat, icoor] = geom[iat, icoor] + delta
            mol.set_geom_(_geom, unit="Bohr")
            e1, g1 = g_scanner(mol)

            H[i, :] = g1.reshape(-1) - g0

    H = (H + H.T) / twodelta
    return H


def central_differences_hess(mol, g_scanner):
    """Evaluate numerical hessian of the energy using central differences."""
    delta = 1e-4
    fourdelta = 4 * delta
    geom = mol.atom_coords()
    nat = geom.shape[0]
    ndim = 3 * nat
    H = np.zeros([ndim, ndim])

    for iat in range(nat):
        for icoor in range(3):
            i = 3 * iat + icoor
            _geom = geom.copy()
            _geom[iat, icoor] = geom[iat, icoor] + delta
            mol.set_geom_(_geom, unit="Bohr")
            e1, g1 = g_scanner(mol)

            _geom[iat, icoor] = geom[iat, icoor] - delta
            mol.set_geom_(_geom, unit="Bohr")
            e2, g2 = g_scanner(mol)

            H[i, :] = (g1 - g2).reshape(-1)

    H = (H + H.T) / fourdelta
    return H


def filter_hessian(mol, H):
    """Projects translations and rotations out of the hessian."""
    # NOTE: This method can maybe be in pyscf.hessian.thermo
    zero_thres = 1e-5
    coords = mol.atom_coords()
    m = mol.atom_mass_list()
    sm = np.sqrt(m)
    sm3 = np.repeat(sm, 3)
    nat = len(m)

    D = np.zeros([6, 3 * nat])
    D[0], D[1], D[2], D[3], D[4], D[5] = _get_TR(m, coords)
    norm = np.linalg.norm(D, axis=1)
    D = D[norm > zero_thres]
    D /= norm[norm > zero_thres][:, None]

    P = np.identity(D.shape[1]) - D.T @ D
    H = np.einsum("ij, i, j -> ij", H, 1 / sm3, 1 / sm3)
    H = P.T @ H @ P
    H = np.einsum("ij, i, j -> ij", H, sm3, sm3)
    return H


def hess_powell_update(H, dq, dg):
    """Update hessian using Powell rule"""
    Hdq = H.dot(dq)
    dqdq = dq.dot(dq)
    aux = (dg - Hdq) * dq[:, None]
    dH1 = (aux + aux.T) / dqdq
    dH2 = (dq.dot(dg) - dq.dot(Hdq)) * dq[:, None] * dq / dqdq ** 2
    dH = dH1 - dH2
    return dH


def hess_BFGS_update(H, dq, dg):
    """Update hessian using BFGS rule"""
    dH1 = dg[None, :] * dg[:, None] / dq.dot(dg)
    dH2 = H.dot(dq[None, :] * dq[:, None]).dot(H) / dq.dot(H).dot(dq)
    dH = dH1 - dH2  # BFGS update
    return dH


if __name__ == "__main__":
    from pyscf import gto, scf

    test_names = ["H2O", "H2"]
    atoms = [
        "O 0. 0.40846003 0.60306451; H 0. -0.770609 2.20504746; H 0. 2.2518751 1.3492855",
        "H 0 0 0; H 0 0 1.603",
    ]

    for (name, atom) in zip(test_names, atoms):
        mol = gto.M(
            atom=atom,
            basis="minao",
            verbose=0,
            unit="Bohr",
        )
        mf = scf.RHF(mol)
        g_scanner = mf.nuc_grad_method().as_scanner()
        H = numhess(mol, g_scanner)
        eigval = np.linalg.eigvalsh(H)

        H = filter_hessian(mol, H)
        feigval = np.linalg.eigvalsh(H)

        print(f"\n {name} example")
        print("Not filtered; Filtered")
        for ival in range(eigval.shape[0]):
            print(f"{eigval[ival]:2.5E} {feigval[ival]:2.5E}")



#!/usr/bin/env python
# Copyright 2014-2020 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

NAME = 'pyscf-qsdopt'
AUTHOR = 'Pyscf Developer'
AUTHOR_EMAIL = None
DESCRIPTION  = 'PySCF extension module qsdopt'
SO_EXTENSIONS = {
}
DEPENDENCIES = ['pyscf', 'numpy']

#######################################################################
# Unless not working, nothing below needs to be changed.
metadata = globals()
import os
import sys
from setuptools import setup, find_namespace_packages, Extension

topdir = os.path.abspath(os.path.join(__file__, '..'))
modules = find_namespace_packages(include=['pyscf.*'])
def guess_version():
    for module in modules:
        module_path = os.path.join(topdir, *module.split('.'))
        for version_file in ['__init__.py', '_version.py']:
            version_file = os.path.join(module_path, version_file)
            if os.path.exists(version_file):
                with open(version_file, 'r') as f:
                    for line in f.readlines():
                        if line.startswith('__version__'):
                            delim = '"' if '"' in line else "'"
                            return line.split(delim)[1]
    raise ValueError("Version string not found")
if not metadata.get('VERSION', None):
    VERSION = guess_version()

pyscf_lib_dir = os.path.join(topdir, 'pyscf', 'lib')
def make_ext(pkg_name, srcs,
             libraries=[], library_dirs=[pyscf_lib_dir],
             include_dirs=[], extra_compile_flags=[],
             extra_link_flags=[], **kwargs):
    if sys.platform.startswith('darwin'):  # OSX
        from distutils.sysconfig import get_config_vars
        conf_vars = get_config_vars()
        conf_vars['LDSHARED'] = conf_vars['LDSHARED'].replace('-bundle', '-dynamiclib')
        conf_vars['CCSHARED'] = " -dynamiclib"
        conf_vars['EXT_SUFFIX'] = '.dylib'
        soname = pkg_name.split('.')[-1]
        extra_link_flags = extra_link_flags + ['-install_name', f'@loader_path/{soname}.dylib']
        runtime_library_dirs = []
    else:
        extra_compile_flags = extra_compile_flags + ['-fopenmp']
        extra_link_flags = extra_link_flags + ['-fopenmp']
        runtime_library_dirs = ['$ORIGIN', '.']
    os.path.join(topdir, *pkg_name.split('.')[:-1])
    return Extension(pkg_name, srcs,
                     libraries = libraries,
                     library_dirs = library_dirs,
                     include_dirs = include_dirs + library_dirs,
                     extra_compile_args = extra_compile_flags,
                     extra_link_args = extra_link_flags,
                     runtime_library_dirs = runtime_library_dirs,
                     **kwargs)

settings = {
    'name': metadata.get('NAME', None),
    'version': VERSION,
    'description': metadata.get('DESCRIPTION', None),
    'author': metadata.get('AUTHOR', None),
    'author_email': metadata.get('AUTHOR_EMAIL', None),
    'install_requires': metadata.get('DEPENDENCIES', []),
}
if 'SO_EXTENSIONS' in metadata:
    settings['ext_modules'] = [make_ext(k, v) for k, v in SO_EXTENSIONS.items()]
setup(
    include_package_data=True,
    packages=modules,
    **settings
)
